/*
   Copyright 2009 Timo Schneider <timos@hrz.tu-chemnitz.de>
   
   This file is part of Tgtd.
   
   Tgtd is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   Tgtd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with Tgtd.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Bot.hpp"

void Bot::set_jid_full(std::string jid) {
	this->jid_full = jid;
} 

void Bot::set_password(std::string pwd) {
	this->password = pwd;
}

void Bot::set_storage(Storage* s) {
	this->stor = s;
}

void Bot::start_bot() {
	JID jid( this->jid_full );
	j = new Client( jid, this->password );
	j->registerMessageHandler(this);
	j->registerSubscriptionHandler(this);
	j->registerSubscriptionHandler(this);
	j->registerPresenceHandler(this);
	j->setPresence( PresenceAvailable, 5 );
	LOG4CXX_DEBUG(logger, "Connecting...");
	j->connect();	
}

void Bot::send_todos(Stanza* stanza) {
	
	std::vector<std::string> tasks;
	std::string peer = stanza->from().full();
	
	tasks = stor->get_tasks_for_jid(stanza->from().bare());

	if (tasks.size() == 0) {
		LOG4CXX_INFO(logger, "No active tasks for " + stanza->from().bare());
		return;
	}
	
	std::stringstream sstr;
	sstr << "Your tasks:\n";
	
	for (int cnt = 0; cnt < tasks.size(); cnt++) {
		int num = cnt + 1;
		sstr << "[" << num << "]  " << tasks[cnt] << "\n";
	}
	
	std::string msg = sstr.str();
	std::string::iterator it = msg.end();
	it--;
	msg.erase(it);
 
	Stanza *s = Stanza::createMessageStanza(peer, msg);
	j->send( s );

}

void Bot::handleMessage( Stanza* stanza, MessageSession* session) {

	// write incoming message to debug log
	LOG4CXX_DEBUG(logger, stanza->xml());

	// ignore messages from unauthorized users  
	if (this->stor->check_if_jid_is_registred(stanza->from().bare()) == false) {
		LOG4CXX_DEBUG(logger, "Ignored message (reason: unauthorized user)");
		return;
	}

	// ignore non-chat messages
	if (stanza->subtype() != StanzaMessageChat) {
		LOG4CXX_DEBUG(logger, "Ignored message (reason: non-chat message)");
		return;
	}
 
	std::string peer = stanza->from().full();

	boost::regex list("^\\W*list\\W*$", boost::regex::perl | boost::regex::icase);	
	boost::regex add("^\\W*add\\W+(.+)\\W*$", boost::regex::perl | boost::regex::icase);	
	boost::regex del("^\\W*del\\W+(\\d+)\\W*$", boost::regex::perl | boost::regex::icase);
	boost::cmatch matches;

	if (boost::regex_match(stanza->body().c_str(), matches, list)) {
		send_todos(stanza);
	}
	else if (regex_match(stanza->body().c_str(), matches, add)) {
		std::string user = stanza->from().bare();
		LOG4CXX_INFO(logger, "Adding " + matches[1] + " as a task for " + user);
		this->stor->add_task_for_user(matches[1], user);
	}
	else if (regex_match(stanza->body().c_str(), matches, del)) { 
		std::string user = stanza->from().bare();
		LOG4CXX_INFO(logger, "Deleting task number " + matches[1] + " from " + user);
		std::istringstream isst;
		isst.str(matches[1]);
		int tasknum;
		isst >> tasknum;
		int numitems_in_db = this->stor->get_number_of_tasks_for_jid(user);
		if ((numitems_in_db < tasknum) or (tasknum <= 0)) {
			std::stringstream sstr;
			sstr << numitems_in_db;
			LOG4CXX_WARN(logger,  "Request to delete item " + isst.str() + " ignored, only " + sstr.str() + " items in database");
			Stanza *s = Stanza::createMessageStanza(peer, "No such item");
			j->send( s );
		}
		else {
			this->stor->delete_task_for_user(tasknum, user);
		}
	}
	else {
		LOG4CXX_WARN(logger, "Unknown command: \"" + stanza->body() + "\"");
		Stanza *s = Stanza::createMessageStanza(peer, "Unknown command");
		j->send( s );			
	}
}

void Bot::handleSubscription(Stanza *stanza) {
	
	// ignore subscription requests from unauthorized users
	if (this->stor->check_if_jid_is_registred(stanza->from().bare()) == false) {
		LOG4CXX_INFO(logger, "Ignoring " + stanza->from().bare() + " subscription request (unregistred user)");
		return;
	}

	// acknowledge subscription and subscribe ourself to get presence updates
	std::string peer = stanza->from().full();
	LOG4CXX_INFO(logger, "Got subscription request from " + peer);
	RosterManager *rm = j->rosterManager();
	rm->ackSubscriptionRequest(stanza->from(), true);
	LOG4CXX_INFO(logger, "Acknowledged subscription request from " + peer);
	rm->subscribe(stanza->from(), stanza->from().username());
	LOG4CXX_INFO(logger, "Subscribed to " + peer + "status");

}

void Bot::handlePresence(Stanza *stanza) {
	
	// ignore presence changes from unauthorized users
	if (this->stor->check_if_jid_is_registred(stanza->from().bare()) == false){
		LOG4CXX_INFO(logger, "Ignoring " + stanza->from().bare() + " presence change (unregistred user)");
		return;
	}
	
	// if user was previously "unavailable" and becomes "available", send todos
	
	std::string peer = stanza->from().full();	
	LOG4CXX_INFO(logger, peer + "changed it's presence");
	std::map<std::string, Presence>::iterator it;
	it = this->resource_status.find(peer);
	
	if (it == this->resource_status.end()) {
		// first contact with this user+resource
		this->resource_status.insert( std::make_pair(peer, stanza->presence()) );
		send_todos(stanza);
	}
	else if (((it->second != PresenceAvailable) or (it->second != PresenceChat)) and
			 ((stanza->presence() == PresenceAvailable) or (stanza->presence() == PresenceChat))) {
		// user was "unavailable" before and became available now
		it->second = stanza->presence();
		send_todos(stanza);
	}
	else {
		it->second = stanza->presence();
		 LOG4CXX_INFO(logger, "Not sending todos to " + peer + " because presence change didn't indicate that user became available");
	}

}

