/*
   Copyright 2009 Timo Schneider <timos@hrz.tu-chemnitz.de>
   
   This file is part of Tgtd.
   
   Tgtd is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   Tgtd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with Tgtd.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Storage.hpp"

bool Storage::prepare_sql_query(std::string sql, sqlite3_stmt **stmt) {

	/**
		This is just a wrapper function arround sqlite3_prepare so that we don't
		have to do error checking and reporting each time. It may only be used by
		other Storage::* functions.
	*/
	
	try {
		LOG4CXX_DEBUG(logger, "Preparing SQL query: " + sql);
		int ret = sqlite3_prepare(this->Database, sql.c_str(), -1, stmt, NULL);
		if (ret != SQLITE_OK) throw(1);
		if (*stmt == NULL) throw(2);
	}
	catch (int e) {
		if (e == 1) {
			std::string err(sqlite3_errmsg(this->Database));
			LOG4CXX_ERROR(logger, "SQLite3 couldn't prepare the query, returncode wasn't SQLITE_OK! Error: " + err);
		}
		else if (e == 2) {
			LOG4CXX_ERROR(logger, "SQLite3 couldn't prepare the query, retuned statement handle is a NULL pointer!");
		}
		return false;
	}
	return true;
}
   
int Storage::get_dbuserid_from_jid(std::string jid) {

	/**
		This function returns the id that belongs to the Jabber ID, which is of
		the form user@server. The returned id is a positive integer value. This
		function terminates if the given jid does not exist. So before calling
		this function the user has to make sure the jid belongs to a registred
		user, possibly by using the appropriate function. Note: This is a
		convienience function which should only be used by other Storage
		functions. All public Storage functions used the jid as identifier for
		users.
	*/
	
	int ret = 0;
	std::stringstream sql;
	sqlite3_stmt *stmt;
	 
	sql << "SELECT id FROM authorized_users WHERE jid = ?;";
	if (prepare_sql_query(sql.str(), &stmt)) {
		LOG4CXX_DEBUG(logger, "Binding value to SQL query: " + jid);
		ret = sqlite3_bind_text(stmt, 1, jid.c_str(), -1, SQLITE_TRANSIENT);
		ret = sqlite3_step(stmt);
		if (ret != SQLITE_ROW) {
			std::string err(sqlite3_errmsg(this->Database));
			LOG4CXX_ERROR(logger, "SQLite3 couldn't execute the query, Error: " + err);
		}
		else {
			ret = sqlite3_column_int(stmt, 0);
		}
		sqlite3_finalize(stmt);
	}
	if (ret < 1) {
		LOG4CXX_FATAL(logger, "Userid for " + jid + " could not be retrieved!" );
		LOG4CXX_FATAL(logger, "Terminating immediately for security reasons!");
		std::exit(EXIT_FAILURE);
	}
			
	return ret;	
	
}

Storage::Storage(std::string Path) {
	
	/**
		Initializes the Storage object by opening the database. The path to
		the database is given by the parameter "Path". Upon sucess
		this.Database contains a sqlite3 database handle. In case of failure
		the program is terminated.
	*/

	//TODO: It might be a good idea to add some sanity checking here, for
	//      example we should check if the database contains all required
	//      tables.

	if (sqlite3_open(Path.c_str(), &Database) != SQLITE_OK) {
		std::string err(sqlite3_errmsg(Database));
		LOG4CXX_FATAL(logger, "Couldn't open Database: " + err);
		std::exit(EXIT_FAILURE);
	}
}

Storage::~Storage() {
	
	/**
		Closes the database.
	*/

	sqlite3_close(Database);
	LOG4CXX_INFO(logger, "Closed Database");
}

bool Storage::check_if_jid_is_registred(std::string jid) {
	
	/**
		This function checks if the Jabber ID jid, which should be of the form
		user@server, belongs to a registred user. If so, true is returned, false
		otherwise. This is the _only_ funtion which will not terminate if you
		call it with a jid that doesn't belong to a registred user.
	*/

	int ret = 0;
	sqlite3_stmt *stmt;
	std::stringstream sql;

	sql << "SELECT COUNT(*) FROM authorized_users WHERE jid = ?;";
	
	if (prepare_sql_query(sql.str(), &stmt)) {
		LOG4CXX_DEBUG(logger, "Binding value to SQL query: " + jid);
		int r = sqlite3_bind_text(stmt, 1, jid.c_str(), -1, SQLITE_TRANSIENT);
		if (r != SQLITE_OK) {
			LOG4CXX_ERROR(logger, "SQLite3 couldn't bind the value to the query!");
		}
		if (sqlite3_step(stmt) != SQLITE_ROW) {
			std::string err(sqlite3_errmsg(this->Database));
			LOG4CXX_ERROR(logger, "SQLite3 couldn't execute the query, Error: " + err);
		}
		else {
			ret = sqlite3_column_int(stmt, 0);
		}
		sqlite3_finalize(stmt);
	}
	
	if (ret > 0) return true;
	else return false;
}

int Storage::get_number_of_tasks_for_jid(std::string jid) {

	/**
		This function returns the number of tasks that are stored for the user
		jid (user@server). The user jid has to be a registred user.
	*/
	
	sqlite3_stmt *stmt;
	std::stringstream sql;

	int ret = get_dbuserid_from_jid(jid);

	sql << "SELECT COUNT(*) FROM tasks WHERE user_id = '" << ret << "';";
	if (prepare_sql_query(sql.str(), &stmt)) {
		if (sqlite3_step(stmt) != SQLITE_ROW) {
			ret = -1;
			std::string err(sqlite3_errmsg(this->Database));
			LOG4CXX_ERROR(logger, "SQLite3 couldn't execute the query, Error: " + err);
		}
		else {
			ret = sqlite3_column_int(stmt, 0);
		}
		sqlite3_finalize(stmt);
	}

	return ret;
}

std::vector<std::string> Storage::get_tasks_for_jid(std::string jid) {
	
	/**
		This function returns all tasks for the user jid (user@server) as a
		vector of strings. The order of the tasks is determined by the order
		the tasks had been added. The most recently added task will be the last
		element in the vector. The user jid has to be a registred user.
	*/

	sqlite3_stmt *stmt;
	std::stringstream sql;
	std::vector<std::string> tasks;
	 
	int ret = get_dbuserid_from_jid(jid);
	sql << "SELECT id, task FROM tasks WHERE user_id = '" << ret << "' ORDER BY id;";
	if (prepare_sql_query(sql.str(), &stmt)) {
		while (sqlite3_step(stmt) == SQLITE_ROW) {
			std::string t( (char *) sqlite3_column_text(stmt, 1));
			tasks.push_back(t);	
		}
		sqlite3_finalize(stmt);
	}

	return tasks;
	
}

void Storage::add_task_for_user(std::string task, std::string jid) {

	/**
		This function adds the string task as a new task for the user jid,
		where jid is of the form user@server and must belong to a registred
		user.
	*/

	// We are dealing with string input here, so we have to use sqlite3_bind_*
	// to avoid SQL injection. Note that most other functions that get a jid as
	// string input use the int get_dbuserid_from_jid(std::string jid) function
	// to get the uid, therefore they don't have to deal with SQL-Escaping in
	// that case.
	
	sqlite3_stmt *stmt;
	std::stringstream sql;
	
	int uid = get_dbuserid_from_jid(jid);
	sql << "INSERT INTO tasks (task, user_id) VALUES (?, ?);";
	if (prepare_sql_query(sql.str(), &stmt)) {
		LOG4CXX_DEBUG(logger, "Binding value to SQL query: " + task);
		int ret = sqlite3_bind_text(stmt, 1, task.c_str(), -1, SQLITE_STATIC);
		if (ret != SQLITE_OK) {
			LOG4CXX_ERROR(logger, "SQLite3 couldn't bind the values to the query!");
		}
		LOG4CXX_DEBUG(logger, "Binding value to SQL query: " + uid);
		ret = sqlite3_bind_int(stmt, 2, uid);
		if (ret != SQLITE_OK) {
			LOG4CXX_ERROR(logger, "SQLite3 couldn't bind the values to the query!");
		}
		if (sqlite3_step(stmt) != SQLITE_DONE) {
			std::string err(sqlite3_errmsg(this->Database));
			LOG4CXX_ERROR(logger, "SQLite3 couldn't execute the query, Error: " + err);
		}
		sqlite3_finalize(stmt);
	}
}

void Storage::delete_task_for_user(int tasknum, std::string jid) {
	
	sqlite3_stmt *stmt;
	std::stringstream sql;
	
	/*
	 At first we have to find out which task was presented to the user as item tasknum.
	 It is _not_ the id of the task, since id is a autoincrement field.
	*/

	int ret = get_dbuserid_from_jid(jid);
	sql << "SELECT id FROM tasks WHERE user_id = '" << ret << "' ORDER BY id LIMIT 1 OFFSET " << tasknum - 1 << ";";
	if (prepare_sql_query(sql.str(), &stmt)) {
		ret = sqlite3_step(stmt);
		if (ret != SQLITE_ROW) {
			std::string err(sqlite3_errmsg(this->Database));
			LOG4CXX_ERROR(logger, "SQLite3 couldn't execute the query, Error: " + err);	
		}
		else {
			ret = sqlite3_column_int(stmt, 0);
		}
		sqlite3_finalize(stmt);
	}

	/*
	 Now the value of ret is the id of the task we want to delete.
	*/

	sql.str("");
	sql.clear();
	sql << "DELETE FROM tasks WHERE id = '" << ret << "';";
	if (prepare_sql_query(sql.str(), &stmt)) {
		ret = sqlite3_step(stmt);
		if (ret != SQLITE_DONE) {
			std::string err(sqlite3_errmsg(this->Database));
			LOG4CXX_ERROR(logger, "SQLite3 couldn't execute the query, Error: " + err);
		}
		sqlite3_finalize(stmt);
	}

}

